"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SerializationVisitor = void 0;
const Domain_1 = require("@/NCalc/Domain");
class SerializationVisitor extends Domain_1.LogicalExpressionVisitor {
    // @todo Find out how to convert this to JS
    _numberFormatInfo;
    EncapsulateNoValue(expression) {
        if (expression.constructor.name == "ValueExpression") {
            expression.Accept(this);
        }
        else {
            this.Result.push("(");
            expression.Accept(this);
            // trim spaces before adding a closing paren
            while (this.Result[this.Result.length - 1] == ' ')
                this.Result.splice(this.Result.length - 1, 1);
            this.Result.push(") ");
        }
    }
    constructor() {
        super();
        this.Result = [];
        this._numberFormatInfo = {};
    }
    Result = [];
    Visit(arg) {
        const requestType = arg.constructor.name;
        switch (requestType) {
            case "LogicalExpression":
                this.VisitLogical(arg);
                break;
            case "TernaryExpression":
                this.VisitTernary(arg);
                break;
            case "BinaryExpression":
                this.VisitBinary(arg);
                break;
            case "UnaryExpression":
                this.VisitUnary(arg);
                break;
            case "ValueExpression":
                this.VisitValue(arg);
                break;
            case "NcalcFunction":
                this.VisitFunction(arg);
                break;
            case "Identifier":
                this.VisitIdentifier(arg);
                break;
            default:
                throw new Error(`Unknown visit type given, received '${requestType}'`);
        }
    }
    VisitLogical(expression) {
        throw new Error("The method or operation is not implemented.");
    }
    VisitTernary(expression) {
        this.EncapsulateNoValue(expression.LeftExpression);
        this.Result.push("? ");
        this.EncapsulateNoValue(expression.MiddleExpression);
        this.Result.push(": ");
        this.EncapsulateNoValue(expression.RightExpression);
    }
    VisitBinary(expression) {
        this.EncapsulateNoValue(expression.LeftExpression);
        switch (expression.Type) {
            case Domain_1.BinaryExpressionType.And:
                this.Result.push("and ");
                break;
            case Domain_1.BinaryExpressionType.Or:
                this.Result.push("or ");
                break;
            case Domain_1.BinaryExpressionType.Div:
                this.Result.push("/ ");
                break;
            case Domain_1.BinaryExpressionType.Equal:
                this.Result.push("= ");
                break;
            case Domain_1.BinaryExpressionType.Greater:
                this.Result.push("> ");
                break;
            case Domain_1.BinaryExpressionType.GreaterOrEqual:
                this.Result.push(">= ");
                break;
            case Domain_1.BinaryExpressionType.Lesser:
                this.Result.push("< ");
                break;
            case Domain_1.BinaryExpressionType.LesserOrEqual:
                this.Result.push("<= ");
                break;
            case Domain_1.BinaryExpressionType.Minus:
                this.Result.push("- ");
                break;
            case Domain_1.BinaryExpressionType.Modulo:
                this.Result.push("% ");
                break;
            case Domain_1.BinaryExpressionType.NotEqual:
                this.Result.push("!= ");
                break;
            case Domain_1.BinaryExpressionType.Plus:
                this.Result.push("+ ");
                break;
            case Domain_1.BinaryExpressionType.Times:
                this.Result.push("* ");
                break;
            case Domain_1.BinaryExpressionType.BitwiseAnd:
                this.Result.push("& ");
                break;
            case Domain_1.BinaryExpressionType.BitwiseOr:
                this.Result.push("| ");
                break;
            case Domain_1.BinaryExpressionType.BitwiseXOr:
                this.Result.push("~ ");
                break;
            case Domain_1.BinaryExpressionType.LeftShift:
                this.Result.push("<< ");
                break;
            case Domain_1.BinaryExpressionType.RightShift:
                this.Result.push(">> ");
                break;
        }
        this.EncapsulateNoValue(expression.RightExpression);
    }
    VisitUnary(expression) {
        switch (expression.Type) {
            case Domain_1.UnaryExpressionType.Not:
                this.Result.push("!");
                break;
            case Domain_1.UnaryExpressionType.Negate:
                this.Result.push("-");
                break;
            case Domain_1.UnaryExpressionType.BitwiseNot:
                this.Result.push("~");
                break;
        }
        this.EncapsulateNoValue(expression.Expression);
    }
    VisitValue(expression) {
        switch (expression.Type) {
            case Domain_1.ValueType.Boolean:
                // @todo ??
                this.Result.push(expression.Value);
                this.Result.push(" ");
                break;
            case Domain_1.ValueType.DateTime:
                this.Result.push("#");
                this.Result.push(expression.Value.toString());
                this.Result.push("#");
                this.Result.push(" ");
                break;
            case Domain_1.ValueType.Float:
                this.Result.push(parseFloat(expression.Value.toString()).toFixed(2));
                this.Result.push(" ");
                break;
            case Domain_1.ValueType.Integer:
                this.Result.push(expression.Value.toString());
                this.Result.push(" ");
                break;
            case Domain_1.ValueType.String:
                this.Result.push("'");
                this.Result.push(expression.Value.toString());
                this.Result.push("'");
                this.Result.push(" ");
                break;
        }
    }
    VisitFunction(func) {
        this.Result.push(func.Identifier.Name);
        this.Result.push("(");
        for (let i = 0; i < func.Expressions.length; i++) {
            func.Expressions[i].Accept(this);
            if (i < func.Expressions.length - 1) {
                this.Result.splice(this.Result.length - 1, 1);
                this.Result.push(", ");
            }
        }
        // trim spaces before adding a closing paren
        while (this.Result[this.Result.length - 1] == ' ')
            this.Result.splice(this.Result.length - 1, 1);
        this.Result.push(") ");
    }
    VisitIdentifier(parameter) {
        this.Result.push("[");
        this.Result.push(parameter.Name);
        this.Result.push("] ");
    }
}
exports.SerializationVisitor = SerializationVisitor;
