"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ValueType = exports.ValueExpression = exports.EvaluationException = void 0;
const Domain_1 = require("@/NCalc/Domain");
class EvaluationException extends Error {
}
exports.EvaluationException = EvaluationException;
class ValueExpression extends Domain_1.LogicalExpression {
    constructor(value, valueType) {
        super();
        if (value) {
            this.Value = value;
            if (!valueType) {
                const detectedType = typeof value;
                switch (detectedType) {
                    case "number":
                    case "bigint":
                        this.Type = ValueType.Integer;
                        break;
                    case "string":
                        // Attempt to check if it is a date
                        if (new Date(value) !== "Invalid Date" && !isNaN(new Date(value))) {
                            this.Type = ValueType.DateTime;
                        }
                        else {
                            this.Type = ValueType.String;
                        }
                        break;
                    case "boolean":
                        this.Type = ValueType.Boolean;
                    default:
                        throw new EvaluationException("This value could not be handled: " + value);
                }
            }
        }
        if (valueType) {
            this.Type = valueType;
        }
    }
    Value = ""; // @todo also wrong to be undefined
    Type = ValueType.Boolean; // @todo what
    Accept(visitor) {
        visitor.Visit(this);
    }
}
exports.ValueExpression = ValueExpression;
var ValueType;
(function (ValueType) {
    ValueType[ValueType["Integer"] = 0] = "Integer";
    ValueType[ValueType["String"] = 1] = "String";
    ValueType[ValueType["DateTime"] = 2] = "DateTime";
    ValueType[ValueType["Float"] = 3] = "Float";
    ValueType[ValueType["Boolean"] = 4] = "Boolean";
})(ValueType = exports.ValueType || (exports.ValueType = {}));
