"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ValueType = exports.ValueExpression = exports.EvaluationException = exports.UnaryExpressionType = exports.UnaryExpression = exports.TernaryExpression = exports.SerializationVisitor = exports.Identifier = exports.ArgumentException = exports.EvaluationVisitor = exports.BinaryExpressionType = exports.BinaryExpression = exports.NCalcFunction = exports.LogicalExpression = exports.LogicalExpressionVisitor = void 0;
const EvaluationOptions_1 = require("../EvaluationOptions");
const Expression_1 = require("../Expression");
const FunctionArgs_1 = require("../FunctionArgs");
const Numbers_1 = require("../Numbers");
const ParameterArgs_1 = require("../ParameterArgs");
class LogicalExpressionVisitor {
}
exports.LogicalExpressionVisitor = LogicalExpressionVisitor;
class LogicalExpression {
    constructor() { }
    And(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.And, this, new ValueExpression(operand));
        }
        else {
            return new BinaryExpression(BinaryExpressionType.And, this, operand);
        }
    }
    DividedBy(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.Div, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.Div, this, new ValueExpression(operand));
        }
    }
    EqualsTo(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.Equal, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.Equal, this, new ValueExpression(operand));
        }
    }
    GreaterThan(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.Greater, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.Greater, this, new ValueExpression(operand));
        }
    }
    GreaterOrEqualThan(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.GreaterOrEqual, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.GreaterOrEqual, this, new ValueExpression(operand));
        }
    }
    LesserThan(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.Lesser, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.Lesser, this, new ValueExpression(operand));
        }
    }
    LesserOrEqualThan(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.LesserOrEqual, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.LesserOrEqual, this, new ValueExpression(operand));
        }
    }
    Minus(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.Minus, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.Minus, this, new ValueExpression(operand));
        }
    }
    Modulo(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.Modulo, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.Modulo, this, new ValueExpression(operand));
        }
    }
    NotEqual(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.NotEqual, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.NotEqual, this, new ValueExpression(operand));
        }
    }
    Or(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.Or, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.Or, this, new ValueExpression(operand));
        }
    }
    Plus(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.Plus, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.Plus, this, new ValueExpression(operand));
        }
    }
    Mult(operand) {
        return this.Plus(operand);
    }
    BitwiseOr(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.BitwiseOr, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.BitwiseOr, this, new ValueExpression(operand));
        }
    }
    BitwiseAnd(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.BitwiseAnd, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.BitwiseAnd, this, new ValueExpression(operand));
        }
    }
    BitwiseXOr(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.BitwiseXOr, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.BitwiseXOr, this, new ValueExpression(operand));
        }
    }
    LeftShift(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.LeftShift, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.LeftShift, this, new ValueExpression(operand));
        }
    }
    RightShift(operand) {
        if (Object.getPrototypeOf(operand) === 'LogicalExpression') {
            return new BinaryExpression(BinaryExpressionType.RightShift, this, operand);
        }
        else {
            return new BinaryExpression(BinaryExpressionType.RightShift, this, new ValueExpression(operand));
        }
    }
    ToString() {
        const serializer = new SerializationVisitor();
        this.Accept(serializer);
        return serializer.Result.join('').trim();
    }
    Accept(visitor) {
        visitor.Visit(this);
    }
}
exports.LogicalExpression = LogicalExpression;
class NCalcFunction extends LogicalExpression {
    constructor(identifier, expressions) {
        super();
        this.Identifier = identifier;
        this.Expressions = expressions;
    }
    Identifier;
    Expressions;
    Accept(visitor) {
        visitor.Visit(this);
    }
}
exports.NCalcFunction = NCalcFunction;
class BinaryExpression extends LogicalExpression {
    constructor(type, leftExpression, rightExpression) {
        super();
        this.Type = type;
        this.LeftExpression = leftExpression;
        this.RightExpression = rightExpression;
    }
    LeftExpression;
    RightExpression;
    Type;
    Accept(visitor) {
        visitor.Visit(this);
    }
}
exports.BinaryExpression = BinaryExpression;
var BinaryExpressionType;
(function (BinaryExpressionType) {
    BinaryExpressionType[BinaryExpressionType["None"] = 0] = "None";
    BinaryExpressionType[BinaryExpressionType["And"] = 1] = "And";
    BinaryExpressionType[BinaryExpressionType["Or"] = 2] = "Or";
    BinaryExpressionType[BinaryExpressionType["NotEqual"] = 3] = "NotEqual";
    BinaryExpressionType[BinaryExpressionType["LesserOrEqual"] = 4] = "LesserOrEqual";
    BinaryExpressionType[BinaryExpressionType["GreaterOrEqual"] = 5] = "GreaterOrEqual";
    BinaryExpressionType[BinaryExpressionType["Lesser"] = 6] = "Lesser";
    BinaryExpressionType[BinaryExpressionType["Greater"] = 7] = "Greater";
    BinaryExpressionType[BinaryExpressionType["Equal"] = 8] = "Equal";
    BinaryExpressionType[BinaryExpressionType["Minus"] = 9] = "Minus";
    BinaryExpressionType[BinaryExpressionType["Plus"] = 10] = "Plus";
    BinaryExpressionType[BinaryExpressionType["Modulo"] = 11] = "Modulo";
    BinaryExpressionType[BinaryExpressionType["Div"] = 12] = "Div";
    BinaryExpressionType[BinaryExpressionType["Times"] = 13] = "Times";
    BinaryExpressionType[BinaryExpressionType["BitwiseOr"] = 14] = "BitwiseOr";
    BinaryExpressionType[BinaryExpressionType["BitwiseAnd"] = 15] = "BitwiseAnd";
    BinaryExpressionType[BinaryExpressionType["BitwiseXOr"] = 16] = "BitwiseXOr";
    BinaryExpressionType[BinaryExpressionType["LeftShift"] = 17] = "LeftShift";
    BinaryExpressionType[BinaryExpressionType["RightShift"] = 18] = "RightShift";
    BinaryExpressionType[BinaryExpressionType["Unknown"] = 19] = "Unknown";
    BinaryExpressionType[BinaryExpressionType["Exponentiation"] = 20] = "Exponentiation";
})(BinaryExpressionType = exports.BinaryExpressionType || (exports.BinaryExpressionType = {}));
class EvaluationVisitor extends LogicalExpressionVisitor {
    // private delegate T Func<T>();
    _options = EvaluationOptions_1.EvaluateOptions.None;
    // private readonly CultureInfo _cultureInfo;
    get IgnoreCase() {
        return this._options & EvaluationOptions_1.EvaluateOptions.IgnoreCase;
    }
    constructor(options) {
        super();
        this._options = options;
    }
    Result;
    Evaluate(expression) {
        expression.Accept(this);
        return this.Result;
    }
    Visit(expression) {
        switch (expression.constructor.name) {
            case 'LogicalExpression':
                throw new Error();
                break;
            case 'TernaryExpression':
                this.VisitTernary(expression);
                break;
            case 'BinaryExpression':
                this.VisitBinaryExpression(expression);
                break;
            case 'UnaryExpression':
                this.VisitUnaryExpression(expression);
                break;
            case 'ValueExpression':
                this.VisitValueExpression(expression);
                break;
            case 'NCalcFunction':
                this.VisitNCalcFunction(expression);
                break;
            case 'Identifier':
                this.VisitIdentifier(expression);
                break;
            default:
                throw new Error('Invalud expression type');
        }
    }
    VisitTernary(expression) {
        // Evaluates the left expression and saves the value
        expression.LeftExpression.Accept(this);
        let left = this.Result == true;
        if (left) {
            expression.MiddleExpression.Accept(this);
        }
        else {
            expression.RightExpression.Accept(this);
        }
    }
    static CommonTypes = ['number', 'boolean', 'string', 'bigint'];
    /// <summary>
    /// Gets the the most precise type.
    /// </summary>
    /// <param name="a">Type a.</param>
    /// <param name="b">Type b.</param>
    /// <returns></returns>
    static GetMostPreciseType(a, b) {
        for (let t in this.CommonTypes) {
            if (typeof a == t || typeof b == t) {
                return t;
            }
        }
        return a;
    }
    // @todo Revisit this function for equality
    CompareUsingMostPreciseType(a, b) {
        if (a == null && b == null) {
            return 0;
        }
        if (a == null) {
            return -1;
        }
        if (b == null) {
            return 1;
        }
        if (typeof a == 'number' || typeof b == 'number') {
            if (a < b) {
                return -1;
            }
            else if (a > b) {
                return 1;
            }
            return 0;
        }
        else {
            return a == b ? 0 : 1;
        }
        // return Comparer.Default.Compare(Convert.ChangeType(a, mpt), Convert.ChangeType(b, mpt));
    }
    static IsReal(value) {
        var typeCode = typeof value;
        return typeCode == Numbers_1.TypeCode.Decimal;
    }
    VisitBinaryExpression(expression) {
        // simulate Lazy<Func<>> behavior for late evaluation
        let leftValue = null;
        const left = () => {
            if (leftValue == null) {
                expression.LeftExpression.Accept(this);
                leftValue = this.Result;
            }
            return leftValue;
        };
        // simulate Lazy<Func<>> behavior for late evaluations
        let rightValue = null;
        const right = () => {
            if (rightValue == null) {
                expression.RightExpression.Accept(this);
                rightValue = this.Result;
            }
            return rightValue;
        };
        // @todo re-evaluate this implementation
        switch (expression.Type) {
            case BinaryExpressionType.And:
                this.Result = left() == true && right() == true;
                break;
            case BinaryExpressionType.Or:
                this.Result = left() == true || right() == true;
                break;
            case BinaryExpressionType.Div:
                this.Result =
                    EvaluationVisitor.IsReal(left()) || EvaluationVisitor.IsReal(right())
                        ? Numbers_1.Numbers.Divide(left(), right())
                        : Numbers_1.Numbers.Divide(parseFloat(left()), right());
                break;
            case BinaryExpressionType.Equal:
                // Use the type of the left operand to make the comparison
                this.Result = this.CompareUsingMostPreciseType(left(), right()) == 0;
                break;
            case BinaryExpressionType.Greater:
                // Use the type of the left operand to make the comparison
                this.Result = this.CompareUsingMostPreciseType(left(), right()) > 0;
                break;
            case BinaryExpressionType.GreaterOrEqual:
                // Use the type of the left operand to make the comparison
                this.Result = this.CompareUsingMostPreciseType(left(), right()) >= 0;
                break;
            case BinaryExpressionType.Lesser:
                // Use the type of the left operand to make the comparison
                this.Result = this.CompareUsingMostPreciseType(left(), right()) < 0;
                break;
            case BinaryExpressionType.LesserOrEqual:
                // Use the type of the left operand to make the comparison
                this.Result = this.CompareUsingMostPreciseType(left(), right()) <= 0;
                break;
            case BinaryExpressionType.Minus:
                this.Result = Numbers_1.Numbers.Subtract(left(), right());
                break;
            case BinaryExpressionType.Modulo:
                this.Result = Numbers_1.Numbers.Modulo(left(), right());
                break;
            case BinaryExpressionType.NotEqual:
                // Use the type of the left operand to make the comparison
                this.Result = this.CompareUsingMostPreciseType(left(), right()) != 0;
                break;
            case BinaryExpressionType.Plus:
                if (typeof left() == 'string') {
                    this.Result = left().concat(right());
                }
                else {
                    this.Result = Numbers_1.Numbers.Add(left(), right());
                }
                break;
            case BinaryExpressionType.Times:
                this.Result = Numbers_1.Numbers.Multiply(left(), right());
                break;
            case BinaryExpressionType.BitwiseAnd:
                this.Result = parseInt(left()) & parseInt(right());
                break;
            case BinaryExpressionType.BitwiseOr:
                this.Result = parseInt(left()) | parseInt(right());
                break;
            case BinaryExpressionType.BitwiseXOr:
                this.Result = parseInt(left()) ^ parseInt(right());
                break;
            case BinaryExpressionType.LeftShift:
                this.Result = parseInt(left()) << parseInt(right());
                break;
            case BinaryExpressionType.RightShift:
                this.Result = parseInt(left()) >> parseInt(right());
                break;
            case BinaryExpressionType.Exponentiation:
                this.Result = Math.pow(parseFloat(left()), parseFloat(right()));
                break;
        }
    }
    VisitUnaryExpression(expression) {
        // Recursively evaluates the underlying expression
        expression.Expression.Accept(this);
        switch (expression.Type) {
            case UnaryExpressionType.Not:
                this.Result = !(this.Result == true);
                break;
            case UnaryExpressionType.Negate:
                this.Result = Numbers_1.Numbers.Subtract(0, this.Result);
                break;
            case UnaryExpressionType.BitwiseNot:
                this.Result = ~parseInt(this.Result);
                break;
            case UnaryExpressionType.Positive:
                // No-op
                break;
        }
    }
    VisitValueExpression(expression) {
        this.Result = expression.Value;
    }
    VisitNCalcFunction(func) {
        let args = new FunctionArgs_1.FunctionArgs();
        args.Parameters = [];
        // Don't call parameters right now, instead let the func do it as needed.
        // Some parameters shouldn't be called, for instance, in a if(), the "not" value might be a division by zero
        // Evaluating every value could produce unexpected behaviour
        for (let i = 0; i < func.Expressions.length; i++) {
            args.Parameters[i] = new Expression_1.Expression(func.Expressions[i], this._options);
            args.Parameters[i].EvaluateFunction = this.EvaluateFunction;
            args.Parameters[i].EvaluateParameter = this.EvaluateParameter;
            // Assign the parameters of the Expression to the arguments so that custom funcs and Parameters can use them
            args.Parameters[i].Parameters = this.Parameters;
        }
        // Calls external implementation
        this.OnEvaluateFunction(this.IgnoreCase ? func.Identifier.Name.toLowerCase() : func.Identifier.Name, args);
        // // If an external implementation was found get the result back
        if (args.HasResult) {
            this.Result = args.Result;
            return;
        }
        switch (func.Identifier.Name.toLowerCase()) {
            // Start Abs
            case 'abs':
                this.CheckCase('Abs', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Abs() takes exactly 1 argument');
                this.Result = Math.abs(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // Start Acos
            case 'acos':
                this.CheckCase('Acos', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Acos() takes exactly 1 argument');
                this.Result = Math.acos(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // Start Asin
            case 'asin':
                this.CheckCase('Asin', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Asin() takes exactly 1 argument');
                this.Result = Math.asin(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // Start Atan
            case 'atan':
                this.CheckCase('Atan', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Atan() takes exactly 1 argument');
                this.Result = Math.atan(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // Start Atan2
            case 'atan2':
                this.CheckCase('Atan2', func.Identifier.Name);
                if (func.Expressions.length != 2)
                    throw new ArgumentException('Atan2() takes exactly 2 argument');
                this.Result = Math.atan2(parseFloat(this.Evaluate(func.Expressions[0])), parseFloat(this.Evaluate(func.Expressions[1])));
                break;
            // end
            // Start Ceiling
            case 'ceiling':
                this.CheckCase('Ceiling', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Ceiling() takes exactly 1 argument');
                this.Result = Math.ceil(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // Start Cos
            case 'cos':
                this.CheckCase('Cos', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Cos() takes exactly 1 argument');
                this.Result = Math.cos(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // Start Exp
            case 'exp':
                this.CheckCase('Exp', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Exp() takes exactly 1 argument');
                this.Result = Math.exp(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // Start Floor
            case 'floor':
                this.CheckCase('Floor', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Floor() takes exactly 1 argument');
                this.Result = Math.floor(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // // Start IEEERemainder
            // case "ieeeremainder":
            //     this.CheckCase("IEEERemainder", func.Identifier.Name);
            //     if (func.Expressions.length != 2)
            //         throw new ArgumentException("IEEERemainder() takes exactly 2 arguments");
            //     Result = Math.IEEERemainder(parseFloat(this.Evaluate(func.Expressions[0])), parseFloat(this.Evaluate(func.Expressions[1])));
            //     break;
            // // end
            // Start Ln
            case 'ln':
                this.CheckCase('Ln', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Ln() takes exactly 1 argument');
                this.Result = Math.log(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // Start Log
            case 'log':
                this.CheckCase('Log', func.Identifier.Name);
                if (func.Expressions.length != 2)
                    throw new ArgumentException('Log() takes exactly 2 arguments');
                this.Result =
                    Math.log(parseFloat(this.Evaluate(func.Expressions[0]))) /
                        Math.log(parseFloat(this.Evaluate(func.Expressions[1])));
                break;
            // end
            // Start Log10
            case 'log10':
                this.CheckCase('Log10', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Log10() takes exactly 1 argument');
                this.Result = Math.log10(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // Start Pow
            case 'pow':
                this.CheckCase('Pow', func.Identifier.Name);
                if (func.Expressions.length != 2)
                    throw new ArgumentException('Pow() takes exactly 2 arguments');
                this.Result = Math.pow(parseFloat(this.Evaluate(func.Expressions[0])), parseFloat(this.Evaluate(func.Expressions[1])));
                break;
            // end
            // Start Round
            // @todo Implementation is incorrect!
            case 'round':
                this.CheckCase('Round', func.Identifier.Name);
                if (func.Expressions.length != 2)
                    throw new ArgumentException('Round() takes exactly 2 arguments');
                // const rounding = (this._options & EvaluateOptions.RoundAwayFromZero) == EvaluateOptions.RoundAwayFromZero ? MidpointRounding.AwayFromZero : MidpointRounding.ToEven;
                this.Result = Math.round(parseFloat(this.Evaluate(func.Expressions[0]))).toFixed(this.Evaluate(func.Expressions[1]));
                break;
            // end
            // Start Sign
            case 'sign':
                this.CheckCase('Sign', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Sign() takes exactly 1 argument');
                this.Result = Math.sign(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // Start Sin
            case 'sin':
                this.CheckCase('Sin', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Sin() takes exactly 1 argument');
                this.Result = Math.sin(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // Start Sqrt
            case 'sqrt':
                this.CheckCase('Sqrt', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Sqrt() takes exactly 1 argument');
                this.Result = Math.sqrt(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // Start Tan
            case 'tan':
                this.CheckCase('Tan', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Tan() takes exactly 1 argument');
                this.Result = Math.tan(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // Start Truncate
            case 'truncate':
                this.CheckCase('Truncate', func.Identifier.Name);
                if (func.Expressions.length != 1)
                    throw new ArgumentException('Truncate() takes exactly 1 argument');
                this.Result = Math.trunc(parseFloat(this.Evaluate(func.Expressions[0])));
                break;
            // end
            // Start Max
            case 'max':
                this.CheckCase('Max', func.Identifier.Name);
                if (func.Expressions.length != 2)
                    throw new ArgumentException('Max() takes exactly 2 arguments');
                const maxleft = this.Evaluate(func.Expressions[0]);
                const maxright = this.Evaluate(func.Expressions[1]);
                this.Result = Numbers_1.Numbers.Max(maxleft, maxright);
                break;
            // end
            // Start Min
            case 'min':
                this.CheckCase('Min', func.Identifier.Name);
                if (func.Expressions.length != 2)
                    throw new ArgumentException('Min() takes exactly 2 arguments');
                const minleft = this.Evaluate(func.Expressions[0]);
                const minright = this.Evaluate(func.Expressions[1]);
                this.Result = Numbers_1.Numbers.Min(minleft, minright);
                break;
            // end
            // Start if
            case 'if':
                this.CheckCase('if', func.Identifier.Name);
                if (func.Expressions.length != 3)
                    throw new ArgumentException('if() takes exactly 3 arguments');
                const val = this.Evaluate(func.Expressions[0]);
                const cond = val == true ? true : false;
                this.Result = cond
                    ? this.Evaluate(func.Expressions[1])
                    : this.Evaluate(func.Expressions[2]);
                break;
            // end
            // Start in
            case 'in':
                this.CheckCase('in', func.Identifier.Name);
                if (func.Expressions.length < 2)
                    throw new ArgumentException('in() takes at least 2 arguments');
                const parameter = this.Evaluate(func.Expressions[0]);
                let evaluation = false;
                // Goes through any values, and stop whe one is found
                for (let i = 1; i < func.Expressions.length; i++) {
                    let argument = this.Evaluate(func.Expressions[i]);
                    if (this.CompareUsingMostPreciseType(parameter, argument) == 0) {
                        evaluation = true;
                        break;
                    }
                }
                this.Result = evaluation;
                break;
            // end
            default:
                throw new ArgumentException(`Function ${func.Identifier.Name} was not found.`);
        }
    }
    CheckCase(func, called) {
        if (this.IgnoreCase) {
            if (equalsIgnoringCase(func, called)) {
                return;
            }
            throw new ArgumentException(`Function '${called}' not found`);
        }
        if (func != called) {
            throw new Error(`Function not found ${called}. Try ${func} instead.`);
        }
    }
    Parameters = {};
    EvaluateFunction = {};
    EvaluateParameter = {};
    OnEvaluateFunction(name, args) {
        if (this.EvaluateFunction.hasOwnProperty(name)) {
            // @todo
            this.EvaluateFunction[name](args);
        }
    }
    OnEvaluateParameter(name, args) {
        if (this.EvaluateParameter.hasOwnProperty(name)) {
            this.EvaluateParameter[name](args);
        }
    }
    VisitIdentifier(parameter) {
        if (this.Parameters.hasOwnProperty(parameter.Name)) {
            // The parameter is defined in the hashtable
            if (this.Parameters[parameter.Name].constructor.name == 'Expression') {
                // The parameter is itself another Expression
                let expression = this.Parameters[parameter.Name];
                // Overloads parameters
                for (let p in this.Parameters) {
                    expression.Parameters[p] = this.Parameters[p];
                }
                expression.EvaluateFunction = this.EvaluateFunction;
                expression.EvaluateParameter = this.EvaluateParameter;
                this.Result = this.Parameters[parameter.Name].Evaluate();
            }
            else {
                this.Result = this.Parameters[parameter.Name];
            }
        }
        else {
            // The parameter should be defined in a call back method
            var args = new ParameterArgs_1.ParameterArgs();
            // Calls external implementation
            this.OnEvaluateParameter(parameter.Name, args);
            if (!args.HasResult)
                throw new ArgumentException(`Parameter '${parameter.Name}' was not defined `);
            this.Result = args.Result;
        }
    }
}
exports.EvaluationVisitor = EvaluationVisitor;
const equalsIgnoringCase = (text, other) => {
    return text.localeCompare(other, undefined, { sensitivity: 'base' }) === 0;
};
class ArgumentException extends Error {
}
exports.ArgumentException = ArgumentException;
class Identifier extends LogicalExpression {
    constructor(name) {
        super();
        this.Name = name;
    }
    Name = '';
    Accept(visitor) {
        visitor.Visit(this);
    }
}
exports.Identifier = Identifier;
class SerializationVisitor extends LogicalExpressionVisitor {
    // @todo Find out how to convert this to JS
    _numberFormatInfo;
    EncapsulateNoValue(expression) {
        if (expression.constructor.name == 'ValueExpression') {
            expression.Accept(this);
        }
        else {
            this.Result.push('(');
            expression.Accept(this);
            // trim spaces before adding a closing paren
            while (this.Result[this.Result.length - 1] == ' ')
                this.Result.splice(this.Result.length - 1, 1);
            this.Result.push(') ');
        }
    }
    constructor() {
        super();
        this.Result = [];
        this._numberFormatInfo = {};
    }
    Result = [];
    Visit(arg) {
        const requestType = arg.constructor.name;
        switch (requestType) {
            case 'LogicalExpression':
                this.VisitLogical(arg);
                break;
            case 'TernaryExpression':
                this.VisitTernary(arg);
                break;
            case 'BinaryExpression':
                this.VisitBinary(arg);
                break;
            case 'UnaryExpression':
                this.VisitUnary(arg);
                break;
            case 'ValueExpression':
                this.VisitValue(arg);
                break;
            case 'NCalcFunction':
                this.VisitFunction(arg);
                break;
            case 'Identifier':
                this.VisitIdentifier(arg);
                break;
            default:
                throw new Error(`Unknown visit type given, received '${requestType}'`);
        }
    }
    VisitLogical(expression) {
        throw new Error('The method or operation is not implemented.');
    }
    VisitTernary(expression) {
        this.EncapsulateNoValue(expression.LeftExpression);
        this.Result.push('? ');
        this.EncapsulateNoValue(expression.MiddleExpression);
        this.Result.push(': ');
        this.EncapsulateNoValue(expression.RightExpression);
    }
    VisitBinary(expression) {
        this.EncapsulateNoValue(expression.LeftExpression);
        switch (expression.Type) {
            case BinaryExpressionType.And:
                this.Result.push('and ');
                break;
            case BinaryExpressionType.Or:
                this.Result.push('or ');
                break;
            case BinaryExpressionType.Div:
                this.Result.push('/ ');
                break;
            case BinaryExpressionType.Equal:
                this.Result.push('= ');
                break;
            case BinaryExpressionType.Greater:
                this.Result.push('> ');
                break;
            case BinaryExpressionType.GreaterOrEqual:
                this.Result.push('>= ');
                break;
            case BinaryExpressionType.Lesser:
                this.Result.push('< ');
                break;
            case BinaryExpressionType.LesserOrEqual:
                this.Result.push('<= ');
                break;
            case BinaryExpressionType.Minus:
                this.Result.push('- ');
                break;
            case BinaryExpressionType.Modulo:
                this.Result.push('% ');
                break;
            case BinaryExpressionType.NotEqual:
                this.Result.push('!= ');
                break;
            case BinaryExpressionType.Plus:
                this.Result.push('+ ');
                break;
            case BinaryExpressionType.Times:
                this.Result.push('* ');
                break;
            case BinaryExpressionType.BitwiseAnd:
                this.Result.push('& ');
                break;
            case BinaryExpressionType.BitwiseOr:
                this.Result.push('| ');
                break;
            case BinaryExpressionType.BitwiseXOr:
                this.Result.push('~ ');
                break;
            case BinaryExpressionType.LeftShift:
                this.Result.push('<< ');
                break;
            case BinaryExpressionType.RightShift:
                this.Result.push('>> ');
                break;
        }
        this.EncapsulateNoValue(expression.RightExpression);
    }
    VisitUnary(expression) {
        switch (expression.Type) {
            case UnaryExpressionType.Not:
                this.Result.push('!');
                break;
            case UnaryExpressionType.Negate:
                this.Result.push('-');
                break;
            case UnaryExpressionType.BitwiseNot:
                this.Result.push('~');
                break;
        }
        this.EncapsulateNoValue(expression.Expression);
    }
    VisitValue(expression) {
        switch (expression.Type) {
            case ValueType.Boolean:
                // @todo ??
                this.Result.push(expression.Value);
                this.Result.push(' ');
                break;
            case ValueType.DateTime:
                this.Result.push('#');
                this.Result.push(expression.Value.toString());
                this.Result.push('#');
                this.Result.push(' ');
                break;
            case ValueType.Float:
                this.Result.push(parseFloat(expression.Value.toString()).toFixed(2));
                this.Result.push(' ');
                break;
            case ValueType.Integer:
                this.Result.push(expression.Value.toString());
                this.Result.push(' ');
                break;
            case ValueType.String:
                this.Result.push("'");
                this.Result.push(expression.Value.toString());
                this.Result.push("'");
                this.Result.push(' ');
                break;
        }
    }
    VisitFunction(func) {
        this.Result.push(func.Identifier.Name);
        this.Result.push('(');
        for (let i = 0; i < func.Expressions.length; i++) {
            func.Expressions[i].Accept(this);
            if (i < func.Expressions.length - 1) {
                this.Result.splice(this.Result.length - 1, 1);
                this.Result.push(', ');
            }
        }
        // trim spaces before adding a closing paren
        while (this.Result[this.Result.length - 1] == ' ')
            this.Result.splice(this.Result.length - 1, 1);
        this.Result.push(') ');
    }
    VisitIdentifier(parameter) {
        this.Result.push('[');
        this.Result.push(parameter.Name);
        this.Result.push('] ');
    }
}
exports.SerializationVisitor = SerializationVisitor;
class TernaryExpression extends LogicalExpression {
    constructor(leftExpression, middleExpression, rightExpression) {
        super();
        this.LeftExpression = leftExpression;
        this.MiddleExpression = middleExpression;
        this.RightExpression = rightExpression;
    }
    LeftExpression;
    MiddleExpression;
    RightExpression;
    Accept(visitor) {
        visitor.Visit(this);
    }
}
exports.TernaryExpression = TernaryExpression;
class UnaryExpression extends LogicalExpression {
    constructor(type, expression) {
        super();
        this.Type = type;
        this.Expression = expression;
    }
    Expression;
    Type;
    Accept(visitor) {
        visitor.Visit(this);
    }
}
exports.UnaryExpression = UnaryExpression;
var UnaryExpressionType;
(function (UnaryExpressionType) {
    UnaryExpressionType[UnaryExpressionType["Not"] = 0] = "Not";
    UnaryExpressionType[UnaryExpressionType["Negate"] = 1] = "Negate";
    UnaryExpressionType[UnaryExpressionType["BitwiseNot"] = 2] = "BitwiseNot";
    UnaryExpressionType[UnaryExpressionType["Positive"] = 3] = "Positive";
})(UnaryExpressionType = exports.UnaryExpressionType || (exports.UnaryExpressionType = {}));
class EvaluationException extends Error {
}
exports.EvaluationException = EvaluationException;
class ValueExpression extends LogicalExpression {
    constructor(value, valueType) {
        super();
        if (value !== null && value != undefined) {
            this.Value = value;
            if (valueType === null || valueType === undefined) {
                const detectedType = typeof value;
                switch (detectedType) {
                    case 'number':
                    case 'bigint':
                        this.Type = ValueType.Integer;
                        break;
                    case 'string':
                        // Attempt to check if it is a date
                        if (new Date(value) !== 'Invalid Date' &&
                            !isNaN(new Date(value))) {
                            this.Type = ValueType.DateTime;
                        }
                        else {
                            this.Type = ValueType.String;
                        }
                        break;
                    case 'boolean':
                        this.Type = ValueType.Boolean;
                        break;
                    default:
                        throw new EvaluationException('This value could not be handled: ' + value);
                }
            }
        }
        if (valueType) {
            this.Type = valueType;
        }
    }
    Value = '';
    Type = ValueType.Boolean;
    Accept(visitor) {
        visitor.Visit(this);
    }
}
exports.ValueExpression = ValueExpression;
var ValueType;
(function (ValueType) {
    ValueType[ValueType["None"] = 0] = "None";
    ValueType[ValueType["Integer"] = 1] = "Integer";
    ValueType[ValueType["String"] = 2] = "String";
    ValueType[ValueType["DateTime"] = 3] = "DateTime";
    ValueType[ValueType["Float"] = 4] = "Float";
    ValueType[ValueType["Boolean"] = 5] = "Boolean";
})(ValueType = exports.ValueType || (exports.ValueType = {}));
