"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Numbers = exports.TypeCode = exports.InvalidOperationException = void 0;
class InvalidOperationException extends Error {
}
exports.InvalidOperationException = InvalidOperationException;
// @todo remove me
var TypeCode;
(function (TypeCode) {
    TypeCode["Boolean"] = "boolean";
    TypeCode["Byte"] = "number";
    TypeCode["SByte"] = "number";
    TypeCode["Int16"] = "number";
    TypeCode["UInt16"] = "number";
    TypeCode["Int32"] = "number";
    TypeCode["UInt32"] = "number";
    TypeCode["Int64"] = "bigint";
    TypeCode["UInt64"] = "bigint";
    TypeCode["Single"] = "number";
    TypeCode["Double"] = "number";
    TypeCode["Decimal"] = "number";
})(TypeCode = exports.TypeCode || (exports.TypeCode = {}));
class Numbers {
    static ConvertIfString(s) {
        if (typeof s === 'string' || s instanceof String) {
            return parseFloat(s);
        }
        return s;
    }
    static Add(a, b) {
        const aValue = Numbers.ConvertIfString(a);
        const bValue = Numbers.ConvertIfString(b);
        const allowedTypes = ['number', 'bigint'];
        // Heree we have to consider the JS magic to be a bit different.
        const typeCodeA = typeof aValue;
        const typeCodeB = typeof bValue;
        if (allowedTypes.indexOf(typeCodeA) == -1 || allowedTypes.indexOf(typeCodeB) == -1) {
            throw new InvalidOperationException(`Operator '+' cannot be applied to operands of type ${typeCodeA} and ${typeCodeB}.`);
        }
        return aValue + bValue;
    }
    static Subtract(a, b) {
        const aValue = Numbers.ConvertIfString(a);
        const bValue = Numbers.ConvertIfString(b);
        const allowedTypes = ['number', 'bigint'];
        // Heree we have to consider the JS magic to be a bit different.
        const typeCodeA = typeof a;
        const typeCodeB = typeof b;
        if (allowedTypes.indexOf(typeCodeA) == -1 || allowedTypes.indexOf(typeCodeB) == -1) {
            throw new InvalidOperationException(`Operator '-' cannot be applied to operands of type ${typeCodeA} and ${typeCodeB}.`);
        }
        return aValue - bValue;
    }
    static Multiply(a, b) {
        const aValue = Numbers.ConvertIfString(a);
        const bValue = Numbers.ConvertIfString(b);
        const allowedTypes = ['number', 'bigint'];
        // Heree we have to consider the JS magic to be a bit different.
        const typeCodeA = typeof a;
        const typeCodeB = typeof b;
        if (allowedTypes.indexOf(typeCodeA) == -1 || allowedTypes.indexOf(typeCodeB) == -1) {
            throw new InvalidOperationException(`Operator '*' cannot be applied to operands of type ${typeCodeA} and ${typeCodeB}.`);
        }
        return aValue * bValue;
    }
    static Divide(a, b) {
        const aValue = Numbers.ConvertIfString(a);
        const bValue = Numbers.ConvertIfString(b);
        const allowedTypes = ['number', 'bigint'];
        // Heree we have to consider the JS magic to be a bit different.
        const typeCodeA = typeof a;
        const typeCodeB = typeof b;
        if (allowedTypes.indexOf(typeCodeA) == -1 || allowedTypes.indexOf(typeCodeB) == -1) {
            throw new InvalidOperationException(`Operator '/' cannot be applied to operands of type ${typeCodeA} and ${typeCodeB}.`);
        }
        return aValue / bValue;
    }
    static Modulo(a, b) {
        const aValue = Numbers.ConvertIfString(a);
        const bValue = Numbers.ConvertIfString(b);
        const allowedTypes = ['number', 'bigint'];
        // Heree we have to consider the JS magic to be a bit different.
        const typeCodeA = typeof a;
        const typeCodeB = typeof b;
        if (allowedTypes.indexOf(typeCodeA) == -1 || allowedTypes.indexOf(typeCodeB) == -1) {
            throw new InvalidOperationException(`Operator '/' cannot be applied to operands of type ${typeCodeA} and ${typeCodeB}.`);
        }
        return aValue % bValue;
    }
    static Max(a, b) {
        const aValue = Numbers.ConvertIfString(a);
        const bValue = Numbers.ConvertIfString(b);
        if (aValue == null && b == null) {
            return null;
        }
        if (aValue == null) {
            return bValue;
        }
        if (bValue == null) {
            return aValue;
        }
        const allowedTypes = ['number', 'bigint'];
        // Heree we have to consider the JS magic to be a bit different.
        const typeCodeA = typeof a;
        const typeCodeB = typeof b;
        if (allowedTypes.indexOf(typeCodeA) == -1 || allowedTypes.indexOf(typeCodeB) == -1) {
            throw new InvalidOperationException(`Operator 'MAX' cannot be applied to operands of type ${typeCodeA} and ${typeCodeB}.`);
        }
        return Math.max(aValue, bValue);
    }
    static Min(a, b) {
        const aValue = Numbers.ConvertIfString(a);
        const bValue = Numbers.ConvertIfString(b);
        if (aValue == null && b == null) {
            return null;
        }
        if (aValue == null) {
            return bValue;
        }
        if (bValue == null) {
            return aValue;
        }
        const allowedTypes = ['number', 'bigint'];
        // Heree we have to consider the JS magic to be a bit different.
        const typeCodeA = typeof a;
        const typeCodeB = typeof b;
        if (allowedTypes.indexOf(typeCodeA) == -1 || allowedTypes.indexOf(typeCodeB) == -1) {
            throw new InvalidOperationException(`Operator 'MIN' cannot be applied to operands of type ${typeCodeA} and ${typeCodeB}.`);
        }
        return Math.min(aValue, bValue);
    }
}
exports.Numbers = Numbers;
